package ru.amse.pomelov.starbugs.client.gui.frames;

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Dimension;
import java.awt.Point;
import java.awt.event.MouseEvent;
import java.awt.event.MouseWheelEvent;
import java.awt.event.MouseWheelListener;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;
import java.rmi.RemoteException;
import java.util.EnumSet;
import java.util.List;

import javax.swing.BorderFactory;
import javax.swing.JFrame;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.JTextField;
import javax.swing.JViewport;
import javax.swing.ScrollPaneConstants;
import javax.swing.border.EmptyBorder;
import javax.swing.event.MouseInputListener;

import exceptions.ClientException;

import ru.amse.pomelov.starbugs.client.ClientDispatcher;
import ru.amse.pomelov.starbugs.client.gui.ControlPanel;
import ru.amse.pomelov.starbugs.client.gui.MiniMapView;
import ru.amse.pomelov.starbugs.client.gui.PlayFieldView;
import ru.amse.pomelov.starbugs.config.Options;
import ru.amse.pomelov.starbugs.units.AbstractUnit;
import ru.amse.pomelov.starbugs.units.Buildings;
import ru.amse.pomelov.starbugs.units.Units;
import ru.amse.pomelov.starbugs.util.FieldPosition;

public class GameFrame extends JFrame {

	private class MouseScrollListener implements MouseInputListener {

		private final JViewport myViewport;
		private Integer x = null;
		private Integer y = null;
		
		public MouseScrollListener(JViewport viewport) {
			myViewport = viewport;
		}
		
		public void mouseClicked(MouseEvent e) {}
		public void mouseEntered(MouseEvent e) {}
		public void mouseExited(MouseEvent e) {}
		
		public void mousePressed(MouseEvent e) {
			if (e.getButton() != MouseEvent.BUTTON3) {
				return;
			}
			x = e.getX();
			y = e.getY();
		}

		public void mouseReleased(MouseEvent e) {
			x = null;
			y = null;
		}

		public void mouseDragged(MouseEvent e) {
			if ((x != null) && (y != null)) {
				Point p = myViewport.getViewPosition();
				int newX = p.x + x - e.getX();
				int newY = p.y + y - e.getY();
				Dimension d = myViewport.getSize();
				int w = d.width;
				int h = d.height;
				Dimension compSize = myViewport.getView().getPreferredSize();
				newX = (newX + w <= compSize.width) ? newX : compSize.width - w;
				newY = (newY + h <= compSize.height) ? newY : compSize.height - h;
				newX = (newX >= 0) ? newX : 0; 
				newY = (newY >= 0) ? newY : 0;
				myViewport.setViewPosition(new Point(newX, newY));
			}
		}
		public void mouseMoved(MouseEvent e) {}
	}
	
	private class MouseZoomListener implements MouseWheelListener {
		public void mouseWheelMoved(MouseWheelEvent e) {
			if (myGameFieldView == null) {
				return;
			}
			int cellSize = myGameFieldView.getCellSize(); 
			cellSize -= 4 * e.getWheelRotation();
			cellSize = (cellSize < 5) ? 5 : cellSize;
			myGameFieldView.setCellSize(cellSize);
			myGameFieldView.updatePreferredSize();
			myMiniMapView.repaint();
		}
	}
	
	
	private static final long serialVersionUID = 1L;

	private PlayFieldView myGameFieldView;
	private MiniMapView myMiniMapView;
	private final ClientDispatcher myClientDispatcher;
	private boolean inGame = true;
	private boolean myTurn = false;
	private JScrollPane scrollPane = new JScrollPane();
	private void createFieldViews() {
		
		scrollPane.setHorizontalScrollBarPolicy(ScrollPaneConstants.HORIZONTAL_SCROLLBAR_ALWAYS);
		scrollPane.setVerticalScrollBarPolicy(ScrollPaneConstants.VERTICAL_SCROLLBAR_ALWAYS);
		myGameFieldView = new PlayFieldView(myClientDispatcher, scrollPane.getViewport());
		
		scrollPane.setViewportView(myGameFieldView);

		
		scrollPane.setWheelScrollingEnabled(false);
		scrollPane.getViewport().setScrollMode(JViewport.BLIT_SCROLL_MODE);
		scrollPane.setPreferredSize(new Dimension(727, 727));
		
		MouseInputListener listener = new MouseScrollListener(scrollPane.getViewport());
		myGameFieldView.addMouseListener(listener);
		myGameFieldView.addMouseMotionListener(listener);
		myGameFieldView.addMouseWheelListener(new MouseZoomListener());
		gamePanel.removeAll();
		gamePanel.add(scrollPane, BorderLayout.CENTER);
		
		myMiniMapView = new MiniMapView(myClientDispatcher, scrollPane.getViewport());
		myMiniMapView.repaint();
		miniMapPanel.removeAll();
		miniMapPanel.add(myMiniMapView, BorderLayout.CENTER);
		pack();
	}
	
	private final JPanel gamePanel = new JPanel(new BorderLayout());
	
	private final JPanel miniMapPanel = new JPanel(new BorderLayout());
	private final ControlPanel myControlPanel;

	private UserWaiting myUserWaitingFrame;

	private final JTextField myTimeLeft = new JTextField();
	
	private class MyWindowListener extends WindowAdapter {
		@Override
		public void windowClosing(WindowEvent e) {
			if (!inGame) {
				myClientDispatcher.forceSurrender();
			} else {
				if (JOptionPane.showConfirmDialog(GameFrame.this, "Closing the game " +
						"will be represented as surrender.\nAre you sure?", "Confirm",
						JOptionPane.YES_NO_OPTION) == JOptionPane.YES_OPTION) {
					if (myTurn) {
						try {
							myClientDispatcher.surrender();
						} catch (ClientException e1) {
							e1.showMessage(GameFrame.this);
						}
					} else {
						myClientDispatcher.forceSurrender();
					}
				}
			}
			
		}
	}
	
	public GameFrame(ClientDispatcher clientDispatcher, boolean monitorOnly) {
		super("StarBugs -- " + (monitorOnly ? "monitor" : "game"));
		setLocation(320, 200);
		setDefaultCloseOperation(DO_NOTHING_ON_CLOSE);
		addWindowListener(new MyWindowListener());
		
		myClientDispatcher = clientDispatcher;
		int max = 0;
		int cur = 0;
		try {
			max = myClientDispatcher.getGameServer().getProperties().playerNum;
			cur = myClientDispatcher.getGameServer().getRegistredPlayerNum();
		} catch (RemoteException e) {
		}

		
		miniMapPanel.setBorder(BorderFactory.createTitledBorder("Mini map"));
		
		getContentPane().setLayout(new BorderLayout());
		
		JPanel mainPanel = new JPanel(new BorderLayout());
		
		mainPanel.add(gamePanel, BorderLayout.CENTER);
		
		JPanel leftPanel = new JPanel(new BorderLayout());
		if (monitorOnly) {
			myControlPanel = null;
		} else {
			myControlPanel = new ControlPanel(myClientDispatcher);
			leftPanel.add(myControlPanel, BorderLayout.CENTER);
		} 
		leftPanel.add(miniMapPanel, BorderLayout.SOUTH);
		
		JPanel leftPanel2 = new JPanel(new BorderLayout());
		leftPanel2.add(leftPanel, BorderLayout.CENTER);

		if (!monitorOnly) {
			myTimeLeft.setEditable(false);
			myTimeLeft.setBorder(new EmptyBorder(3, 10, 3, 10));
			myTimeLeft.setHorizontalAlignment(JTextField.CENTER);
			myTimeLeft.setText("waiting for opponents");
			myTimeLeft.setBackground(new Color(238, 238, 238));
			myTimeLeft.setForeground(Color.BLACK);
			mainPanel.add(myTimeLeft, BorderLayout.NORTH);
		}
		
		mainPanel.add(leftPanel2, BorderLayout.EAST);
		getContentPane().add(mainPanel, BorderLayout.CENTER);
		
		setSize(959, 803);
		
		myUserWaitingFrame = new UserWaiting(max, cur);
		if ((max == cur) && monitorOnly) {
			gameStarted();
		}
	}
	
	private int timeLeft;
	public void decreaseTime(int d) {
		timeLeft -= d;
		myTimeLeft.setText("Your turn! Time left: " + timeLeft);
		if (timeLeft == 0) {
			myTimeLeft.setBackground(Color.RED);
			myTimeLeft.setText("Time is up! Press 'End of turn'");
		}
	}
	public int getTimeLeft() {
		return timeLeft;
	}
	
	public void prepareForTurn(EnumSet<Units> allowedUnits, EnumSet<Buildings> allowedBuildings, int timeAvailable) {
		myTimeLeft.setBackground(Color.GREEN);
		myTimeLeft.setForeground(Color.WHITE);
		timeLeft = timeAvailable;
		myTimeLeft.setText("Your turn! Time left: " + timeAvailable);
		myControlPanel.enableAllowed(allowedUnits, allowedBuildings);
		myTurn = true;
	}
	
	public void endOfTurn() {
		myTurn = false;
		myTimeLeft.setBackground(new Color(238, 238, 238));
		myTimeLeft.setForeground(Color.BLACK);
		myTimeLeft.setText("Waiting for opponents");	
	}
	
	public void surrender() {
		myTimeLeft.setBackground(new Color(238, 238, 238));
		myTimeLeft.setForeground(Color.BLACK);
		myTimeLeft.setText("you are out");
		inGame = false;
	}
	
	public void createdUnits(List<AbstractUnit> unitList) {
		myGameFieldView.putUnits(unitList);
		myGameFieldView.repaint();
		myMiniMapView.putUnits(unitList);
		myMiniMapView.repaint();		
	}
	
	public void removeUnits(List<FieldPosition> coordsList) {
		for (FieldPosition coords : coordsList) {
			myGameFieldView.removeUnit(coords);
			myMiniMapView.removeUnit(coords);
		}
		myGameFieldView.repaint();
		myMiniMapView.repaint();
	}
	
	public void gameStarted() {
		createFieldViews();
		scrollPane.setBorder(BorderFactory.createLineBorder(
				Options.getColor(myClientDispatcher.getMyNum()), 7));
		setVisible(true);
		myUserWaitingFrame.dispose();
	}
	
	public void playerJioned() {
		myUserWaitingFrame.playerJoined();
	}
	
	public void disposeUserWaiting() {
		myUserWaitingFrame.dispose();
	}
}
