package ru.amse.pomelov.starbugs.server;

import java.io.IOException;
import java.io.InputStream;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.OutputStream;
import java.rmi.RemoteException;
import java.util.ArrayList;
import java.util.EnumSet;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Queue;
import java.util.Set;

import ru.amse.pomelov.starbugs.config.Options;
import ru.amse.pomelov.starbugs.remote.GameServer;
import ru.amse.pomelov.starbugs.remote.RemoteGameServerListener;
import ru.amse.pomelov.starbugs.server.positioners.PerimeterBasePositioner;
import ru.amse.pomelov.starbugs.server.positioners.Positioner;
import ru.amse.pomelov.starbugs.server.testing.Saver;
import ru.amse.pomelov.starbugs.turns.AbstractTurn;
import ru.amse.pomelov.starbugs.units.AbstractUnit;
import ru.amse.pomelov.starbugs.units.AbstractUnitTypes;
import ru.amse.pomelov.starbugs.units.Building;
import ru.amse.pomelov.starbugs.units.Buildings;
import ru.amse.pomelov.starbugs.units.Edifice;
import ru.amse.pomelov.starbugs.units.Edifices;
import ru.amse.pomelov.starbugs.units.Units;
import ru.amse.pomelov.starbugs.util.FieldPosition;
import ru.amse.pomelov.starbugs.util.GameProperties;
import ru.amse.pomelov.starbugs.util.MaskIterator;
import ru.amse.pomelov.starbugs.util.SurroundGenerator;

public final class GameDispatcher {
	private final GameField myGameField;
	
	private final List<PlayerAccount> myPlayers = 
		new ArrayList<PlayerAccount>();

	private final Set<Integer> myOutPlayers = new HashSet<Integer>();
	
	private int myCurrentPlayer = 0;
	private final GameServer myGameServer;
	private final int numOfPlayers;
	private final List<RemoteGameServerListener> myFieldChangeListeners = new LinkedList<RemoteGameServerListener>();
	
	public static int distanse(FieldPosition coords1, FieldPosition coords2) {
		int x1 = coords1.getX();
		int y1 = coords1.getY();
		int x2 = coords2.getX();
		int y2 = coords2.getY();
		return Math.max(Math.abs(x2 - x1), Math.abs(y2 - y1));
	}
	
	public static boolean areNeighbours(FieldPosition coords1, FieldPosition coords2) {
		return distanse(coords1, coords2) == 1;
	}
	
	public GameDispatcher(GameProperties props, GameServer gameServer, RemoteGameServerListener l) {
		myGameField = new GameField(props.width, props.height);
		numOfPlayers = props.playerNum;
		myGameServer = gameServer;
		
		if (l != null) {
			myFieldChangeListeners.add(l);
		}
		
		Positioner bp = new PerimeterBasePositioner(numOfPlayers, props.width, props.height);
		
		for (int i = 0; i < numOfPlayers; i++) {
			PlayerAccount p = new PlayerAccount(Options.maxTurnTime, i + 1);
			
			LinkedList<Edifice> baseList = new LinkedList<Edifice>();
			FieldPosition basePosition = bp.next();
			Edifice base = new Edifice(basePosition, Edifices.BASE_EDIFICE);
			for (FieldPosition c : new MaskIterator(Edifices.BASE_EDIFICE.getMask(), basePosition)) {
				myGameField.putUnit(new Building(c, i + 1, Buildings.BASE));
			}
			
			while (!base.proceed()) {}
			
			baseList.add(base);
			p.myEdifices.put(Edifices.BASE_EDIFICE.getBuildingType(), baseList);
			myPlayers.add(p);
		}
		
	}
	
	public GameDispatcher(GameField gf) {
		myGameField = gf;
		numOfPlayers = 1;
		myGameServer = null;
	}
	
	private GameDispatcher(int num, int cur, List<PlayerAccount> players, GameField gf, GameServer gs) {
		numOfPlayers = num;
		myCurrentPlayer = cur;
		for (PlayerAccount p : players) {
			myPlayers.add(p);
		}
		myGameField = gf;
		myGameServer = gs;
	}
	
	public GameField getGameField() {
		return myGameField;
	}
	
	/**
	 * Gets mask of building connections to specified cell.
	 * @param p player.
	 * @param x X-coordinate of checking cell.
	 * @param y Y-coordinate of checking cell.
	 * @return connection mask.
	 */
	public Set<Buildings> buildingConnectedMask(int player, FieldPosition coords) {
		Queue<FieldPosition> queue = new LinkedList<FieldPosition>();
		Set<FieldPosition> flags = new HashSet<FieldPosition>();

		queue.add(coords);
		flags.add(coords);
		
		Set<Buildings> result = new HashSet<Buildings>();
		
		while (!queue.isEmpty()) {
			SurroundGenerator sg = new SurroundGenerator(queue.poll());
			
			while (sg.hasNext()) {
				FieldPosition tmp = sg.next();
				if (!flags.contains(tmp) && myGameField.checkCoordinates(tmp)) {
					AbstractUnit u = myGameField.getUnit(tmp);
					flags.add(tmp);
					if ((u != null) && (u.getPlayer() == player)) {
						if (u.getType() == AbstractUnitTypes.BUILDING) {
							result.add(((Building)u).getBuildingType());
						}
						queue.add(tmp);
					}
				}
			}
		}
		
		return result;
	}
	
	
	public boolean isConnectedToBuildings(int player, Set<Buildings> buildingMask, FieldPosition coords) {
		Set<Buildings> connectedMask = buildingConnectedMask(player, coords);
		return connectedMask.containsAll(buildingMask);
	}
	
	public void addBuildingSite(FieldPosition coords, int player, Edifice e) {
		if ((coords == null) || (e == null)) {
			throw new IllegalArgumentException();
		}
		
		PlayerAccount p = myPlayers.get(player - 1);
		p.myBuildingSites.add(e);
	}
	
	
	
	public void nextPlayer() {
		myPlayers.get(myCurrentPlayer).endOfTurn(this);
		do {
			myCurrentPlayer = (myCurrentPlayer + 1) % numOfPlayers;
		} while (myOutPlayers.contains(myCurrentPlayer));
		
		myPlayers.get(myCurrentPlayer).updateEdifices(this);
		
		if (myGameServer != null) {
			try {
				myGameServer.nextPlayer(myCurrentPlayer, myPlayers.get(myCurrentPlayer).getTurnTimeLeft());
			} catch (RemoteException e) {
				surrender();
			}
		}
	}
	
	/**
	 * Checks if the cell is active. In fact if the cell is connected to the base.
	 * @param p player
	 * @param x X-coordinate of checking cell
	 * @param y Y-coordinate of checking cell
	 * @return true if connected
	 */
	public boolean isActive(int player, FieldPosition coords) {
		return isConnectedToBuildings(player, EnumSet.of(Buildings.BASE), coords);
	}
	
	public boolean isEmptyCell(FieldPosition coords) {
		return myGameField.getUnit(coords) == null;
	}
	
	public int cellOwner(FieldPosition coords) {
		AbstractUnit u = myGameField.getUnit(coords);
		return (u == null) ? 0 : u.getPlayer();
	}
	
	public void putUnits(List<AbstractUnit> unitList) {
		for (RemoteGameServerListener l : myFieldChangeListeners) {
			try {
				l.createUnits(unitList);
			} catch (RemoteException e) {
			}
		}
		for (AbstractUnit unit : unitList) {
			myGameField.putUnit(unit);
		}
	}
	
	public void removeUnits(List<FieldPosition> unitList) {
		for (RemoteGameServerListener l : myFieldChangeListeners) {
			try {
				l.removeUnits(unitList);
			} catch (RemoteException e) {
			}
		}
		for (FieldPosition unit : unitList) {
			myGameField.removeUnit(unit);
		}		
	}
	
	private void notifyAllGameEnd(int winner) {
		for (RemoteGameServerListener l : myFieldChangeListeners) {
			try {
				l.endOfGame(winner);
			} catch (RemoteException e) {
			}
		}		
	}
	
	public AbstractUnit getUnit(FieldPosition coords) {
		return myGameField.getUnit(coords);
	}
	
	public int proceedTurn(AbstractTurn turn) {
		if (turn == null) {
			return -1;
		}
		
		if ((turn.getPlayer() == -1) || (myCurrentPlayer + 1 != turn.getPlayer())) {
			return -1;
		}
		
		int d = turn.acceptTurn(this);
		if (d == -1) {
			return -1;
		}
		PlayerAccount player = myPlayers.get(myCurrentPlayer);
		if (d > player.getTurnTimeLeft()) {
			return -1;
		} else {
			player.decreaseTime(d);
			turn.makeTurn(this);
			return d;
		}
	}
	
	public void SaveGame(OutputStream out) throws IOException{
		ObjectOutputStream oos = new ObjectOutputStream(out);
		oos.writeInt(numOfPlayers);
		oos.writeInt(myCurrentPlayer);
		for (PlayerAccount p : myPlayers) {
			oos.writeObject(p);
		}
		myGameField.accept(new Saver(out));
	}

	public static GameDispatcher LoadGame(InputStream in, GameServer gameServer) throws IOException {
		ObjectInputStream ois = new ObjectInputStream(in);
		int num = ois.readInt();
		int cur = ois.readInt();
		List<PlayerAccount> players = new ArrayList<PlayerAccount>();
		for (int i = 0; i < num; i++) {
			try {
				PlayerAccount p = (PlayerAccount) ois.readObject();
				players.add(p);
			} catch (ClassNotFoundException e) {
				System.out.println("Loading error");
			}
		}
		GameField gf = GameField.load(in);
		return new GameDispatcher(num, cur, players, gf, gameServer);
	}

	public void surrender() {
		List <FieldPosition> toRemove = new LinkedList<FieldPosition>();
		List <AbstractUnit> toNeutral = new LinkedList<AbstractUnit>();
		for (int i = 0; i < myGameField.getWidth(); i++) {
			for (int j = 0; j < myGameField.getHeight(); j++) {
				AbstractUnit u = myGameField.getUnit(FieldPosition.createPoint(i, j));
				if (u != null) {
					if (u.getPlayer() == myCurrentPlayer + 1) {
						if ((u.getType() == AbstractUnitTypes.HUMAN)
							|| (u.getType() == AbstractUnitTypes.MACHINE)) {
							toRemove.add(u.getPosition());
						} else {
							u.setPlayer(0);
							toNeutral.add(u);
						}
					}
				}
			}
		}
		removeUnits(toRemove);
		putUnits(toNeutral);
		
		myOutPlayers.add(myCurrentPlayer);
		
		if (myOutPlayers.size() + 1 == numOfPlayers) {
			for (int i = 0; i < numOfPlayers; i++) {
				if (!myOutPlayers.contains(i)) {
					notifyAllGameEnd(i + 1);
					return;
				}
			}
		}
		
		nextPlayer();
	}
	
	
	public EnumSet<Units> getAllowedUnits() {
		EnumSet<Units> result = EnumSet.noneOf(Units.class);
		
		for (Units u : EnumSet.allOf(Units.class)) {
			Buildings ownBuilding = u.getOwnBuilding();
			List<Edifice> list = myPlayers.get(myCurrentPlayer).myEdifices.get(ownBuilding);
			if ((list != null) && (!list.isEmpty())) {
				result.add(u);
			}
		}
		
		return result;
	}
	
	public EnumSet<Buildings> getAllowedBuildings() {
		EnumSet<Buildings> result = EnumSet.noneOf(Buildings.class);
		
		for (Buildings u : EnumSet.allOf(Buildings.class)) {
			Set<Buildings> dependence = u.getDependence();
			boolean flag = true;
			Map<Buildings, LinkedList<Edifice>> edifices = myPlayers.get(myCurrentPlayer).myEdifices;
			for (Buildings e : EnumSet.allOf(Buildings.class)) {
				LinkedList<Edifice> list = edifices.get(e);
				if (dependence.contains(e) && ((list == null) || (list.isEmpty()))) {
					flag = false;
				}
			}
			if (flag) {
				result.add(u);
			}
		}
		
		return result;
	}
}